#include <stdio.h>
#include <string.h>
#include <ctype.h>
/* #include <dos.h> */
/* #include <dir.h> */
#include <stdlib.h>
#include "intercep.h"

#define MAXFILESIZE 20000		/* maximum template file size */
#define MAXLINELEN 200			/* maximum output line length */
#define AVGLINELEN 40			/* average line length */

char template_text[ MAXFILESIZE ];		/* file text */
char *templates[ MAXFILESIZE/AVGLINELEN ];	/* pointers to each line */
int ntemplates;		/* number of templates */

/* copy template to dest, changing regname to regval throughout.
 * assumes dest is big enough to hold result.
 */
void
subst(char *regname, unsigned rval, char *template, char *dest)
{
	char regval[5];
	if (regname[1] == 'H' || regname[1] == 'L')
		sprintf(regval, "%02X", rval);
	else
		sprintf(regval, "%04X", rval);
	while ( (*dest = *template) ) { /* copy through '\0' at end of string. */
		if (template[0] == regname[0] && template[1] == regname[1]) {
			dest = stpcpy(dest, regval);
			template += 2;	/* all regnames are 2 chars long */
		}
		else {
			template++;
			dest++;
		}
	}
}

int is_a_blank(char c) {
	if (c == ' ')
		return 1;
	if (c == '\t')
		return 1;
	return 0;
}

/* Read template file into buffer, setting pointers to
 * beginning of lines.
 * Sets template_text, templates and ntemplates.
 * Pads out interrupt IDs to 6 characters.
 * Returns number of lines.
 */
int
read_template_file(char *filename)
{
	FILE *ifile;
	char variableinline[ MAXLINELEN ];
	char **nextline = templates;
	char *dest = template_text;
	char *q = template_text;
	int plus, questionmark;

	if (! (ifile = fopen(filename, "rt")))
		return 0;
	while (fgets(variableinline, MAXLINELEN, ifile)) {
		char *p = variableinline;
		plus = 0;
		questionmark = 0;
		while (is_a_blank(*p)) ++p;
		if (*p == '+') {
			++p;
			plus = 1;
		}
		while (is_a_blank(*p)) ++p;
		if (plus && *p == '?') {
			++p;
			questionmark = 1;
		}
		while (is_a_blank(*p)) ++p;
		if (! isxdigit(*p))
			continue;	/* handle comment lines */
		*nextline++ = dest = q;
		while (isxdigit(*p) || (plus && toupper(*p) == 'M'))
			*q++ = toupper(*p++);	/* copy initial hex string */
		while (q < dest+6)
			*q++ = ' ';		/* pad out to 6 characters */
		if (questionmark) {
			*q++ = '?';
		}
		*p = ' ';			/* convert tab (or whatever) to space */
		while ( (*q++ = *p++) )
			;				/* copy rest of line through null */
		while ( (q[-3] == '\\' && q[-2] == '\n')
			|| (q[-4] == '\\' && q[-3] == '\r' && q[-2] == '\n') ) {
			/* continuation lines */
			if (q[-3] == '\r') {
				q -= 4;
			} else {
				q -= 3;
			}
			if ( (p = fgets(variableinline, MAXLINELEN, ifile)) ) {
				*q++ = '\n';
				while ( (*q++ = *p++) )
					;
			}
			else
				goto on_eof;
		}
	}
on_eof:
	*nextline = NULL;
	return (ntemplates = nextline - templates);
}

int compare_templates(const void *v1, const void *v2)
{
	const char ** t1 = (const char **)v1;
	const char ** t2 = (const char **)v2;
	return(strncmp(*t1, *t2, 6));
}

/* find a matching template given interrupt number and ax value
 * works from the most specific (all 6 characters)
 * back to the most general (just the 2 characters of the interrupt number).
 */
char *find_template(int intnum, unsigned axval)
{
	char proposed[ 7 ];
	char *key = proposed;
	char **found;
	char *ignoring = proposed + 6;
	sprintf(proposed, "%02XMM%02X", intnum & 0xFF, axval & 0xFF);
	/* search for AL function with AH=MM first (for AMIS) */
	while (ignoring > proposed + 2 /* do not match "15" without MM */) {
		if ( (found = bsearch(&key, templates, ntemplates,
			sizeof(char *), compare_templates)) )
				return(*found);
		/* else ignore 2 more characters */
		ignoring -= 2;
		ignoring[0] = ignoring[1] = ' ';
	}

	ignoring = proposed + 6;
	sprintf(proposed, "%02X%04X", intnum & 0xFF, axval & 0xFFFF);
	while (ignoring > proposed) {
		if ( (found = bsearch(&key, templates, ntemplates,
			sizeof(char *), compare_templates)) )
				return(*found);
		/* else ignore 2 more characters */
		ignoring -= 2;
		ignoring[0] = ignoring[1] = ' ';
	}
	return NULL;
}

#define HI(u)	(((u)&0xff00)>>8)
#define LO(u)	((u)&0x00ff)

/* read swi_info records from input file
 * interpret them, and output to output file.
 */
void
interpret_file(FILE *ifp, FILE *ofp, char *progname,
	int longmode, int quiet, int format)
{
	swi_info rec;
	swi_info_amis * amis;
	char s1[ MAXLINELEN ];
	char s2[ MAXLINELEN ];
	char *templ;

	if (progname != NULL)
		fprintf (ofp, "INTERCEPT/INTERPRET by Ned Konz, 1987-08-02\n"
			" Extended by C. Masloch, 2021\n"
			"Dump of DOS/BIOS calls from program \"%s\"\n"
			"\n",
			progname);
	if (longmode)
		fprintf(ofp,
			"  CS:IP   INT  AX    BX   CX   DX    DS   ES   SI   DI   BP\n");

	while (fread(&rec, sizeof(swi_info), 1, ifp)) {
		if (format && (rec.intnum & 0xFF00) != 0) {
			switch (rec.intnum) {
			case 0x100:
				amis = (swi_info_amis *)&rec;
				fprintf(ofp, "Multiplexer replied:"
					" %04X:%04X -> \"%8.8s\" \"%8.8s\""
					" version %04X\n",
					amis->segment, amis->offset,
					amis->vendor, amis->product,
					amis->version);
				break;
			default:
				break;
			}
			continue;
		}
		if ( (templ = find_template(rec.intnum, rec.ax)) ) {
			{
				char * src = templ;
				char * dest = s2;
				int ii;
				for (ii = 0; ii < 6; ++ii) {
					*dest++ = *src++;
				}
				if (*src == '?') {
					if (quiet) {
						continue;
					}
					++ src;
				}
				strcpy(dest, src);
			}
			subst("AX", rec.ax,     s2, s1);
			subst("AH", HI(rec.ax), s1, s2);
			subst("AL", LO(rec.ax), s2, s1);

			subst("BX", rec.bx,     s1, s2);
			subst("BH", HI(rec.bx), s2, s1);
			subst("BL", LO(rec.bx), s1, s2);

			subst("CX", rec.cx,     s2, s1);
			subst("CH", HI(rec.cx), s1, s2);
			subst("CL", LO(rec.cx), s2, s1);

			subst("DX", rec.dx,         s1, s2);
			subst("DH", HI(rec.dx),     s2, s1);
			subst("DL", LO(rec.dx),     s1, s2);

			subst("BP", rec.bp,         s2, s1);
			subst("DI", rec.di,         s1, s2);
			subst("SI", rec.si,         s2, s1);
			subst("DS", rec.ds,         s1, s2);
			subst("ES", rec.es,         s2, s1);

			subst("CS", rec.cs,       s1, s2);
			subst("IP", rec.ip,       s2, s1);
			fprintf(ofp, "%04X:%04X  %s",
				rec.cs,
				rec.ip - 2,
				s1);
		}

		if ((templ==NULL) || longmode) {
		  fprintf(ofp,
			"%04X:%04X  %02X %04X  "
			"%04X %04X %04X  "
			"%04X %04X %04X %04X %04X\n",
			rec.cs,
			rec.ip - 2,
			rec.intnum,
			rec.ax, /*AX*/
			rec.bx /*BX*/, rec.cx,/*CX*/ rec.dx,
			rec.ds, rec.es, rec.si, rec.di,
			rec.bp);
		}
	} /*while*/

	return;
}

void usage(const char * progname) {
	fprintf(stderr, "INTERCEPT/INTERPRET by Ned Konz, 1987-08-02\n"
	" Extended by C. Masloch, 2021\n"
	"usage: %s [-l] [-q] [-f V1] [-t templ_file] infile outfile [progname]\n"
	"\t-l\tLong mode (Dump all register values)\n"
	"\t-q\tQuiet mode (Hide templates marked with '?')\n"
	"\t-f V1\tSet format V1 (Extensions, eg replied AMIS signatures)\n"
	"\t-t\tSpecify template data filename\n"
	, progname);
}

int main(int argc, char *argv[])
{
	int longmode = 0;
	int quiet = 0;
	int format = 0;
	char *templfile = TFILENAME;
	char *progname = argv[0];
	FILE *ifp, *ofp;

	if (argc < 3) {
		usage(progname);
		exit(1);
	}
	argv++;

	while (argv[0] != NULL && argv[0][0] == '-') {
		if (! strcasecmp(argv[0], "-l")) { longmode = 1; }
		else if (! strcasecmp(argv[0], "-q")) { quiet = 1; }
		else if (! strcasecmp(argv[0], "-f")) {
			if (! argv[1]) {
				fprintf(stderr, "%s: switch \"%s\" missing parameter\n", progname, argv[0]);
				exit(6);
			}
			if (! strcasecmp(argv[1], "v0")) {
				format = 0;
			} else if (! strcasecmp(argv[1], "v1")) {
				format = 1;
			} else {
				fprintf(stderr, "%s: switch \"%s\" invalid parameter \"%s\"\n", progname, argv[0], argv[1]);
				exit(7);
			}
			++ argv;
		} else if (! strcasecmp(argv[0], "-t")) {
			if (! argv[1]) {
				fprintf(stderr, "%s: switch \"%s\" missing parameter\n", progname, argv[0]);
				exit(6);
			}
			templfile = argv[1];
			argv++;
		} else if (! strcasecmp(argv[0], "-h")
			|| ! strcasecmp(argv[0], "-?")) {
			usage(progname);
			exit(1);
		} else {
			fprintf(stderr, "%s: unknown switch \"%s\"\n", progname, argv[0]);
			exit(4);
		}
		argv++;
	}

	if (argv[0] == NULL || argv[1] == NULL) {
		fprintf(stderr, "%s: missing filenames\n", progname);
		exit(8);
	}
	if (! read_template_file( templfile )) {
		fprintf(stderr, "%s: error during read of \"%s\"\n",
			progname, templfile);
		exit(5);
	}
	qsort(templates, ntemplates, sizeof(char *), compare_templates);
	if (! (ifp = fopen(argv[0], "rb"))) {
		fprintf(stderr, "%s: can't open input file \"%s\"\n",
			progname, argv[0]);
		exit(2);
	}
	if (! (ofp = fopen(argv[1], "wt"))) {
		fprintf(stderr, "%s: can't open output file \"%s\"\n",
			progname, argv[1]);
		exit(3);
	}
	interpret_file(ifp, ofp, argv[2], longmode, quiet, format);
	fclose(ifp);
	fclose(ofp);
	exit(0);
}
